(* $Id: parser.ml,v 1.13 2021/05/11 02:22:20 garrigue Exp $ *)

open Terms

let implode l =
  let len = List.length l in
  let s = Bytes.create len in
  let rec fill n =
    function
      [] -> ()
    | c :: l -> s.[n] <- c; fill (n + 1) l
  in
  fill 0 l; Bytes.to_string s

let rec star ?(acc = []) p (strm__ : _ Stream.t) =
  match try Some (p strm__) with Stream.Failure -> None with
    Some x -> let s = strm__ in star ~acc:(x :: acc) p s
  | _ -> List.rev acc

let rec skip a (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some b -> Stream.junk strm__; let s = strm__ in if a <> b then skip a s
  | _ -> raise Stream.Failure

let alphanum (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some ('A'..'Z' | 'a'..'z' | '0'..'9' | '_' | '-' | '*' | '?' as c) ->
      Stream.junk strm__; c
  | _ -> raise Stream.Failure

let atom (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some ('a'..'z' as c) ->
      Stream.junk strm__;
      let l =
        try star alphanum ~acc:[c] strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      implode l
  | _ -> raise Stream.Failure

let ident (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some ('A'..'Z' | '_' as c) ->
      Stream.junk strm__;
      let l =
        try star alphanum ~acc:[c] strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      implode l
  | _ -> raise Stream.Failure

let num (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some ('0'..'9' as c) -> Stream.junk strm__; c
  | _ -> raise Stream.Failure

let number (strm__ : _ Stream.t) =
  let c = num strm__ in
  let l =
    try star num ~acc:[c] strm__ with
      Stream.Failure -> raise (Stream.Error "")
  in
  int_of_string (implode l)

type token =
    Kwd of string
  | Atm of string
  | Ident of string
  | Num of int
  | EOF

let keywords = ["read"; "write"; "while"; "do"; "define"; "nil"]

let next (strm__ : _ Stream.t) =
  match try Some (atom strm__) with Stream.Failure -> None with
    Some s -> if List.mem s keywords then Kwd s else Atm s
  | _ ->
      match try Some (ident strm__) with Stream.Failure -> None with
        Some s -> Ident s
      | _ ->
          match try Some (number strm__) with Stream.Failure -> None with
            Some n -> Num n
          | _ ->
              match Stream.peek strm__ with
                Some (';' | '(' | ')' | '{' | '}' | '.' as c) ->
                  Stream.junk strm__; Kwd (String.make 1 c)
              | Some ':' ->
                  Stream.junk strm__;
                  begin match Stream.peek strm__ with
                    Some '=' -> Stream.junk strm__; Kwd ":="
                  | _ -> raise (Stream.Error "")
                  end
              | Some '=' ->
                  Stream.junk strm__;
                  begin match Stream.peek strm__ with
                    Some '?' -> Stream.junk strm__; Atm "=?"
                  | _ -> raise (Stream.Error "")
                  end
              | _ -> raise Stream.Failure

let lineno = ref (Stream.sempty, 0, 0)
let get_pos s =
  let (s', n, b) = !lineno in
  if s != s' then 1, Stream.count s else let p = Stream.count s in n, p - b
  
let rec lexer (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some '(' -> Stream.junk strm__; may_comment strm__
  | _ ->
      match try Some (next strm__) with Stream.Failure -> None with
        Some t ->
          let s = strm__ in Stream.icons t (Stream.slazy (fun _ -> lexer s))
      | _ ->
          match Stream.peek strm__ with
            Some (' ' | '\t') -> Stream.junk strm__; lexer strm__
          | Some ('\r' | '\n') ->
              Stream.junk strm__;
              let s = strm__ in
              let p = Stream.count strm__ in
              let (s', n, b) = !lineno in
              if s == s' then lineno := s, n + 1, p else lineno := s, 2, p;
              lexer s
          | Some _ -> Stream.junk strm__; raise (Stream.Error "Lexing")
          | _ -> Stream.ising EOF
and may_comment (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some '*' ->
      Stream.junk strm__;
      let _ =
        try skip '*' strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      may_uncomment strm__
  | _ ->
      let s = strm__ in
      Stream.icons (Kwd "(") (Stream.slazy (fun _ -> lexer s))
and may_uncomment (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some ')' -> Stream.junk strm__; lexer strm__
  | _ -> let _ = skip '*' strm__ in may_uncomment strm__

let rec val_of_num n = if n = 0 then Nil else Cons (Nil, val_of_num (n - 1))

let rec build_list r =
  function
    [] -> r
  | a :: l -> Cons (a, build_list r l)

let rec build_cons r =
  function
    [] -> r
  | a :: l -> Cons (Atom "cons", Cons (a, Cons (build_cons r l, Nil)))

let rec unsnoc acc =
  function
    [b] -> List.rev acc, b
  | [] -> failwith "unsnoc"
  | a :: l -> unsnoc (a :: acc) l
let unsnoc l = unsnoc [] l

let rec expr (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "(") ->
      Stream.junk strm__;
      let l =
        try star expr strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      let e =
        try may_dotexpr strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some (Kwd ")") ->
          Stream.junk strm__;
          begin match l, e with
            Atom "list" :: l, Nil -> build_cons Nil l
          | Atom "cons*" :: (_ :: _ as l), Nil ->
              let (l, e) = unsnoc l in build_cons e l
          | _ -> build_list e l
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (Kwd "nil") -> Stream.junk strm__; Nil
  | Some (Atm a) -> Stream.junk strm__; Atom a
  | Some (Ident x) -> Stream.junk strm__; Var x
  | Some (Num n) -> Stream.junk strm__; val_of_num n
  | _ -> raise Stream.Failure
and may_dotexpr (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd ".") ->
      Stream.junk strm__;
      (try expr strm__ with Stream.Failure -> raise (Stream.Error ""))
  | _ -> Nil

let rec build_command =
  function
    [] -> Skip
  | [c] -> c
  | c :: l -> Seq (c, build_command l)

let build_expr =
  function
    [e] -> e
  | l -> build_list Nil l

let rec commands pre (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd ":=") ->
      Stream.junk strm__;
      let (exprs, cmds, res) =
        try body strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      let (pre', cmd) =
        try
          match unsnoc pre with
            pre', Var x ->
              begin match exprs with
                Atom ("cons" | "hd" | "tl" | "=?") :: _ :: _ ->
                  pre', Assign (x, build_list Nil exprs)
              | Atom f :: exprs ->
                  let (l, r) = try unsnoc exprs with _ -> [], Nil in
                  pre', Call (x, f, build_cons r l)
              | [v] -> pre', Assign (x, v)
              | _ -> failwith "wrong assignment expression"
              end
          | _ -> failwith "unsnoc"
        with Failure "unsnoc" -> raise (Stream.Error "no variable to assign")
      in
      pre', cmd :: cmds, res
  | Some (Kwd ";") ->
      Stream.junk strm__;
      let (cmds, res) =
        try body0 strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      pre, cmds, res
  | Some (Kwd "{") ->
      Stream.junk strm__;
      let (cmds, res) =
        try body0 strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      let (cmds', res') =
        try body0 strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      if res <> "" then failwith "bad command syntax";
      pre, build_command cmds :: cmds', res'
  | Some (Kwd "while") ->
      Stream.junk strm__;
      let es =
        try star expr strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some (Kwd "do") ->
          Stream.junk strm__;
          let (cmds, res) =
            try body0 strm__ with Stream.Failure -> raise (Stream.Error "")
          in
          begin match cmds with
            cmd :: cmds -> pre, While (build_expr es, cmd) :: cmds, res
          | [] -> failwith "missing command after while"
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (Kwd "write") ->
      Stream.junk strm__;
      begin match Stream.peek strm__ with
        Some (Ident y) -> Stream.junk strm__; pre, [], y
      | _ -> raise (Stream.Error "")
      end
  | Some (Kwd "}") -> Stream.junk strm__; pre, [], ""
  | _ -> raise Stream.Failure
and body (strm__ : _ Stream.t) =
  let exprs = star expr strm__ in
  try commands exprs strm__ with Stream.Failure -> raise (Stream.Error "")
and body0 (strm__ : _ Stream.t) =
  let (exprs, cmds, res) = body strm__ in
  if exprs <> [] then failwith "bad command syntax"; cmds, res

let extract_var =
  function
    Var x -> x
  | _ -> failwith "parameter must be variable"

let program (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (Kwd "read") ->
      Stream.junk strm__;
      let (vars, cmds, res) =
        try body strm__ with Stream.Failure -> raise (Stream.Error "")
      in
      if res = "" then failwith "bad program syntax";
      List.map extract_var vars, build_command cmds, res
  | _ -> raise Stream.Failure

(*
let rec expr = parser
    [< e = eval_expr >] -> e
  | [< v = const_elem >] -> Val v
  | [< 'Kwd"("; e = paren_expr; 'Kwd")" >] -> e
and paren_expr = parser
    [< e = eval_expr >] -> e
  | [< v = const_exprs >] -> Val v
and eval_expr = parser
    [< 'Kwd "hd"; e = expr >] -> Ehd e
  | [< 'Kwd "tl"; e = expr >] -> Etl e
  | [< 'Kwd "=?"; e1 = expr; e2 = expr >] -> Eq (e1, e2)
  | [< 'Kwd "cons"; e1 = expr; e2 = expr >] -> Econs (e1, e2)
  | [< 'Kwd "list"; l = star expr >] -> build_cons (Val Nil) l
  | [< 'Kwd "cons*"; l = star expr >] ->
      begin match List.rev l with [] -> Val Nil
      | e::l -> build_cons e (List.rev l)
      end
  | [< 'Ident s >] -> Var s
and const_exprs = parser
    [< l = star const_elem; r = const_end >] -> build_list r l
and const_expr = parser
    [< v = const_elem >] -> v
  | [< 'Kwd"("; v = const_exprs; 'Kwd")" >] -> v
and const_elem = parser
    [< 'Atm s >] -> Atom s
  | [< 'Num n >] -> val_of_num n
  | [< 'Kwd "nil" >] -> Nil
and const_end = parser
    [< 'Kwd"."; v = const_expr >] -> v
  | [< >] -> Nil

let opt_semi =
  parser [< 'Kwd";" >] -> () | [< >] -> ()

let rec commands = parser
    [< l = star command0 >] -> build_command l
and command0 = parser [< c = command; _ = opt_semi >] -> c
and command = parser
    [< 'Ident x; 'Kwd":="; r = assign x >] -> r
  | [< 'Kwd"while"; e = expr; 'Kwd"do" ?? "missing `do' after `while'";
       c = command >] -> While(e,c)
  | [< 'Kwd"{"; c = commands; 'Kwd"}" >] -> c
and assign x = parser
    [< 'Atm f; r = may_call x f >] -> r
  | [< e = expr >] -> Assign(x,e)
and may_call x f = parser
    [< e = expr >] -> Call(x,f,e)
  | [< >] -> Assign(x, Val (Atom f))

let program = parser
    [< 'Atm"read"; 'Ident x; _ = opt_semi; c = commands;
       'Atm"write"; 'Ident y >] -> (x, c, y)
*)

let expr_of_string s = expr (lexer (Stream.of_string s))
let program_of_string s = program (lexer (Stream.of_string s))
let program_of_channel c = program (lexer (Stream.of_channel c))
