(* $Id: parser.ml4,v 1.1 2021/04/27 05:57:09 garrigue Exp $ *)

open Terms

let implode l =
  let len = List.length l in
  let s = Bytes.create len in
  let rec fill n = function
      [] -> ()
    | c::l -> s.[n] <- c; fill (n+1) l
  in
  fill 0 l; Bytes.to_string s

let rec star ?(acc=[]) p = parser
    [< x = p ; s >] -> star ~acc:(x::acc) p s
  | [< >] -> List.rev acc

let rec skip a = parser
    [< 'b; s >] -> if a <> b then skip a s

let alphanum = parser
    [< ' ('A'..'Z'|'a'..'z'|'0'..'9'|'_'|'-'|'*'|'?') as c >] -> c

let atom = parser
    [< ''a'..'z' as c; l = star alphanum ~acc:[c] >] -> implode l

let ident = parser
    [< ' ('A'..'Z'|'_') as c; l = star alphanum ~acc:[c] >] -> implode l

let num = parser [< ''0'..'9' as c >] -> c

let number =
  parser [< c = num; l = star num ~acc:[c] >] -> int_of_string (implode l)

type token =
    Kwd of string | Atm of string | Ident of string | Num of int | EOF

let keywords = ["read";"write";"while";"do";"define";"nil"]

let next = parser
    [< s = atom >] -> if List.mem s keywords then Kwd s else Atm s
  | [< s = ident >] -> Ident s
  | [< n = number >] -> Num n
  | [< ' (';'|'('|')'|'{'|'}'|'.') as c >] -> Kwd (String.make 1 c)
  | [< '':'; ''=' >] -> Kwd ":="
  | [< ''='; ''?' >] -> Atm "=?"

let lineno = ref ([< >],0,0)
let get_pos s =
  let (s',n,b) = !lineno in
  if s != s' then (1, Stream.count s) else
  let p = Stream.count s in
  (n, p-b)
  
let rec lexer = parser
    [< ''('; s >] -> may_comment s
  | [< t = next; s >] -> [< 't; lexer s >]
  | [< ' (' '|'\t'); s >] -> lexer s
  | [< ' ('\r'|'\n'); s >] p ->
      let (s',n,b) = !lineno in
      if s == s' then lineno := (s,n+1,p) else lineno := (s,2,p);
      lexer s
  | [< ' _ >] -> raise (Stream.Error "Lexing")
  | [< >] -> [< 'EOF >]
and may_comment = parser
    [< ''*'; _ = skip '*'; s >] -> may_uncomment s
  | [< s >] -> [< 'Kwd"("; lexer s >]
and may_uncomment = parser
    [< '')'; s >] -> lexer s
  | [< _ = skip '*'; s >] -> may_uncomment s

let rec val_of_num n =
  if n = 0 then Nil else Cons(Nil, val_of_num (n-1))

let rec build_list r = function
    [] -> r
  | a :: l -> Cons (a, build_list r l)

let rec build_cons r = function
    [] -> r
  | a::l -> Cons (Atom"cons", Cons(a, Cons (build_cons r l, Nil)))

let rec unsnoc acc = function
    [b] -> (List.rev acc, b)
  | [] -> failwith "unsnoc"
  | a::l -> unsnoc (a::acc) l
let unsnoc l = unsnoc [] l

let rec expr = parser
    [< 'Kwd"("; l = star expr; e = may_dotexpr; 'Kwd")" >] ->
      begin match (l,e) with
        Atom"list"::l, Nil -> build_cons Nil l
      | Atom"cons*"::(_::_ as l), Nil ->
          let (l,e) = unsnoc l in build_cons e l
      | _ -> build_list e l
      end
  | [< 'Kwd"nil" >] -> Nil
  | [< 'Atm a >] -> Atom a
  | [< 'Ident x >] -> Var x
  | [< 'Num n >] -> val_of_num n
and may_dotexpr = parser
    [< 'Kwd"."; e = expr >] -> e
  | [< >] -> Nil

let rec build_command = function
    [] -> Skip
  | [c] -> c
  | c::l -> Seq(c, build_command l)

let build_expr = function
    [e] -> e
  | l -> build_list Nil l

let rec commands pre = parser
    [< 'Kwd":="; (exprs, cmds, res) = body >] ->
      let (pre', cmd) =
        try match unsnoc pre with
          (pre',Var x) ->
            begin match exprs with
              Atom("cons"|"hd"|"tl"|"=?")::_::_ ->
                (pre', Assign(x, build_list Nil exprs))
            | Atom f :: exprs ->
                let l, r =
                  try unsnoc exprs with _ -> [], Nil
                in
                (pre', Call(x,f,build_cons r l))
            | [v] -> (pre', Assign(x, v))
            | _ -> failwith "wrong assignment expression"
            end
        | _ -> failwith"unsnoc"
        with Failure"unsnoc" ->
          raise (Stream.Error "no variable to assign")
      in (pre', cmd::cmds, res)
  | [< 'Kwd";"; (cmds, res) = body0 >] ->
      (pre, cmds, res)
  | [< 'Kwd"{"; (cmds, res) = body0; (cmds', res') = body0 >] ->
      if res <> "" then failwith "bad command syntax";
      (pre, build_command cmds :: cmds', res')
  | [< 'Kwd"while"; es = star expr; 'Kwd"do"; (cmds, res) = body0 >] ->
      begin match cmds with
        cmd::cmds ->
          (pre, While(build_expr es, cmd)::cmds, res)
      | [] ->
          failwith "missing command after while"
      end
  | [< 'Kwd"write"; 'Ident y >] ->
      (pre, [], y)
  | [< 'Kwd"}" >] ->
      (pre, [], "")
and body = parser
    [< exprs = star expr; r = commands exprs >] -> r
and body0 = parser
    [< (exprs, cmds, res) = body >] ->
      if exprs <> [] then failwith "bad command syntax";
      (cmds, res)

let extract_var = function
  | Var x -> x
  | _ -> failwith "parameter must be variable"

let program = parser
    [< 'Kwd"read"; (vars, cmds, res) = body >] ->
      if res = "" then failwith "bad program syntax";
      (List.map extract_var vars, build_command cmds, res)

(*
let rec expr = parser
    [< e = eval_expr >] -> e
  | [< v = const_elem >] -> Val v
  | [< 'Kwd"("; e = paren_expr; 'Kwd")" >] -> e
and paren_expr = parser
    [< e = eval_expr >] -> e
  | [< v = const_exprs >] -> Val v
and eval_expr = parser
    [< 'Kwd "hd"; e = expr >] -> Ehd e
  | [< 'Kwd "tl"; e = expr >] -> Etl e
  | [< 'Kwd "=?"; e1 = expr; e2 = expr >] -> Eq (e1, e2)
  | [< 'Kwd "cons"; e1 = expr; e2 = expr >] -> Econs (e1, e2)
  | [< 'Kwd "list"; l = star expr >] -> build_cons (Val Nil) l
  | [< 'Kwd "cons*"; l = star expr >] ->
      begin match List.rev l with [] -> Val Nil
      | e::l -> build_cons e (List.rev l)
      end
  | [< 'Ident s >] -> Var s
and const_exprs = parser
    [< l = star const_elem; r = const_end >] -> build_list r l
and const_expr = parser
    [< v = const_elem >] -> v
  | [< 'Kwd"("; v = const_exprs; 'Kwd")" >] -> v
and const_elem = parser
    [< 'Atm s >] -> Atom s
  | [< 'Num n >] -> val_of_num n
  | [< 'Kwd "nil" >] -> Nil
and const_end = parser
    [< 'Kwd"."; v = const_expr >] -> v
  | [< >] -> Nil

let opt_semi =
  parser [< 'Kwd";" >] -> () | [< >] -> ()

let rec commands = parser
    [< l = star command0 >] -> build_command l
and command0 = parser [< c = command; _ = opt_semi >] -> c
and command = parser
    [< 'Ident x; 'Kwd":="; r = assign x >] -> r
  | [< 'Kwd"while"; e = expr; 'Kwd"do" ?? "missing `do' after `while'";
       c = command >] -> While(e,c)
  | [< 'Kwd"{"; c = commands; 'Kwd"}" >] -> c
and assign x = parser
    [< 'Atm f; r = may_call x f >] -> r
  | [< e = expr >] -> Assign(x,e)
and may_call x f = parser
    [< e = expr >] -> Call(x,f,e)
  | [< >] -> Assign(x, Val (Atom f))

let program = parser
    [< 'Atm"read"; 'Ident x; _ = opt_semi; c = commands;
       'Atm"write"; 'Ident y >] -> (x, c, y)
*)

let expr_of_string s = expr (lexer (Stream.of_string s))
let program_of_string s = program (lexer (Stream.of_string s))
let program_of_channel c = program (lexer (Stream.of_channel c))
